//! databd.cpp
//!	a board that stores the position and can compute data about that position.
//
// The contents of this file are subject to
//	the OpenGo Public License Version 2.0 (the "License"); you may not
//	use this file except in compliance with the License. You may obtain
//	a copy of the License at http://www.inventivity.com/OpenGo
//
// Software distributed under the License is distributed on an "AS IS" basis,
//	WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
//	for the specific language governing rights and limitations under
//	the License.
// 
// Copyright:
//      Matthieu Nottale        copyright (c) 2005
//	Jeffrey Greenberg	copyright (c) 1996-2002
// 
// Contributor(s):
//
//==========================================================================


#include "databd.h"

// *****************************************************************************
// *****************************************************************************

GoString::GoString()
{
	_LibertyCount = 0;
}

GoString::~GoString()
{
	// DO NOT TOUCH ANYTHING IN THE CELL!!
	//	Doing so would break code in Absorb() for example...
//	// clear the cell references to this string...
//	ListOf_pCells::iterator ppCell;
//	for( ppCell = _Cells.begin(); ppCell != _Cells.end(); ppCell++ )
//		(*ppCell)->_pInStr = 0;
}

//! put cell into string and point string into cell.
void GoString::AddOccupiedCell( BoardCell *pCell )
{
	Assert( pCell->_piece.GetColor() != PC_None );
	pCell->_pInStr = this;
	_Cells.push_back( pCell );
}

//! put cell into string and point string into cell.
void GoString::AddEmptyCell( BoardCell *pCell )
{
	Assert( pCell->_piece.GetColor() == PC_None );
	pCell->_pInStr = this;
	_Cells.push_back( pCell );
}

void GoString::MarkAll()
{
	ListOf_pCells::iterator ppCell;
	for( ppCell = _Cells.begin(); ppCell != _Cells.end(); ++ppCell )
		(*ppCell)->_piece.SetMark();
}
void GoString::UnMarkAll()
{
	ListOf_pCells::iterator ppCell;
	for( ppCell = _Cells.begin(); ppCell != _Cells.end(); ++ppCell )
		(*ppCell)->_piece.ClearMark();
}


//! Point the other's cells to this string.
void GoString::Absorb( GoString *pOther)
{
	Assert( pOther != this );
	ListOf_pCells::iterator ppCell;

	for( ppCell = pOther->_Cells.begin(); ppCell != pOther->_Cells.end(); ppCell++ )
		AddOccupiedCell( *ppCell );
}

//! Calculates liberties for string.
//! 	Board is marked and cleared in the process...
int GoString::CalcLiberties( DataBoard *pBoard )
{
	// First count the unmarked cells, marking them as we go.
	_LibertyCount = 0;
	ListOf_pCells::iterator ppCell;
	for( ppCell = _Cells.begin(); ppCell != _Cells.end(); ++ppCell )
		_LibertyCount += pBoard->CountUnMarkedLiberties( &(*ppCell)->_piece );

	// Clear the marks.
	for( ppCell = _Cells.begin(); ppCell != _Cells.end(); ++ppCell )
		pBoard->ClearMarkedLiberties( &(*ppCell)->_piece );

	return _LibertyCount;
}


//! copy the pieces out of GoString and into the list...
void GoString::CopyToListAndClear( vector<Piece> * pList ) 
{
	ListOf_pCells::iterator ppCell;
// cout << "DataBoard::CopyToListAndClear ";
	for( ppCell = _Cells.begin(); ppCell != _Cells.end(); ++ppCell )
	{
		Assert( (*ppCell)->_piece.GetColor() != PC_None );
		pList->push_back( (*ppCell)->_piece );
// char buf[64];
// (*ppCell)->_piece.GetString( buf, 64);
// cout <<  buf << ", ";
		(*ppCell)->_piece.SetColor(PC_None);
		(*ppCell)->_pInStr = 0;
	}
// cout << endl;
}

BOOL GoString::Contains(Pt * pt) {
  ListOf_pCells::iterator ppCell;
  for( ppCell = _Cells.begin(); ppCell != _Cells.end(); ++ppCell )
    if ((*ppCell)->_piece.GetCol() == pt->GetCol() &&
	(*ppCell)->_piece.GetRow() == pt->GetRow())
      return TRUE;
  return FALSE;
}


//! Clear the cell's references to the GoString
void GoString::StopCellsFromUsingMe() 
{
	ListOf_pCells::iterator ppCell;
	for( ppCell = _Cells.begin(); ppCell != _Cells.end(); ++ppCell )
		(*ppCell)->_pInStr = 0;
}

#if 0
//! debugging
void GoString::CheckForRepeats()
{
	// check if this string already has a cell in it...
	ListOf_pCells::iterator ppCell;
	for( ppCell = _Cells.begin(); ppCell != _Cells.end(); ++ppCell )
	{
		ListOf_pCells::iterator ppNextCell = ppCell;
		for( ++ppNextCell; ppNextCell != _Cells.end(); ++ppNextCell )
			if( ((*ppCell)->_piece.GetRow() == (*ppNextCell)->_piece.GetRow())
			&&	((*ppCell)->_piece.GetCol() == (*ppNextCell)->_piece.GetCol()) )
				Fatal(ERR_ASSERT);
	}
}

//! debugging
void GoString::Validate() 
{
	ListOf_pCells::iterator ppCell;
// int x = 0;
// cout << "validate: " << _Cells.size() << " entries\n";
	for( ppCell = _Cells.begin(); ppCell != _Cells.end(); ++ppCell )
	{
// char buf[64];
// (*ppCell)->_piece.GetString( buf, 64 );
// cout << x++ << ": " << buf << endl;
		Assert( (*ppCell)->_pInStr == this);
		Assert( (*ppCell)->_piece.GetColor() != PC_None);
	}

	CheckForRepeats();
// cout << "=====" << endl;
}

#endif

// *****************************************************************************
// *****************************************************************************

GoStringCollection::GoStringCollection()
{
	_TotalLiberties = 0;
}

GoStringCollection::~GoStringCollection()
{
	Flush();
}

void GoStringCollection::Add( GoString *pGS )
{
	Assert( pGS );
	_Strings.push_back( pGS );
}

void GoStringCollection::Flush()
{
	_TotalLiberties = 0;
	ListOf_GoStrings::iterator ppGS;
	for( ppGS = _Strings.begin(); ppGS != _Strings.end(); ppGS++ )
	{
		(*ppGS)->StopCellsFromUsingMe();
		delete (*ppGS);
		(*ppGS) = 0;
	}
	_Strings.erase( _Strings.begin(), _Strings.end() );
}

int GoStringCollection::CalcLiberties( DataBoard *pBoard )
{
	_TotalLiberties = 0;
	ListOf_GoStrings::iterator ppGS;
	for( ppGS = _Strings.begin(); ppGS != _Strings.end(); ppGS++ )
		_TotalLiberties += (*ppGS)->CalcLiberties( pBoard );
	return _TotalLiberties;
}

int GoStringCollection::GetLibertyCount()
{
	return _TotalLiberties;
}

GoString * GoStringCollection::Find(Pt *pt) {
  ListOf_GoStrings::iterator ppGS;
  for( ppGS = _Strings.begin(); ppGS != _Strings.end(); ppGS++)
    {
      if ((*ppGS)->Contains(pt))
	return *ppGS;
    }
  return 0;
}


int GoStringCollection::RemoveDead( vector<Piece> * pTheDead )
{
	int AmtKilled = 0;
	ListOf_GoStrings::iterator ppGS;
	for( ppGS = _Strings.begin(); ppGS != _Strings.end(); )
	{
		if( (*ppGS)->GetLibertyCount() == 0)
		{
			++AmtKilled;
			// clear cell's reference to the GoString:
			(*ppGS)->CopyToListAndClear( pTheDead );
			// delete the GoString itself.
			delete (*ppGS);
			*(ppGS) = 0;
			// remove GoString pointer from collection.
			ppGS = _Strings.erase( ppGS );
		}
		else
			++ppGS;
	}
	return AmtKilled;
}

BOOL GoStringCollection::Remove( GoString * pVictim )
{
	ListOf_GoStrings::iterator ppGS;
	for( ppGS = _Strings.begin(); ppGS != _Strings.end(); )
	{
		if( (*ppGS) == pVictim )
		{
			ppGS = _Strings.erase( ppGS );
			return TRUE;	// found;
		}
		else
			++ppGS;
	}
	return FALSE;	// not found
}

// *****************************************************************************
// *****************************************************************************

#define USE_KOCHECK	1	//! check for KO

#include <memory.h>

DataBoard::DataBoard( BOOL bSuicideIsOk, BOOL bSimpleKo, Coord Size, IOBase *pIO )
{
	_bSuicideIsOk = bSuicideIsOk;
	_bSimpleKo = bSimpleKo;
	_maxsize = Size;

	_pIO = pIO;	// for debugging

	// fill the board with nothing i.e. empty pieces.
	for( int row = 0; row < _maxsize; row++ )
		for( int col = 0; col < _maxsize; col++ ) {
			BoardCell *pCell = &_pBoard[row][col];
			pCell->_piece.SetCoords( row, col );
			pCell->_piece.SetColor( PC_None );
            }
}

DataBoard::~DataBoard()
{
	_DumpEnclosures();
}


// -----------------------------------------------------------------

//! Count only unmarked adjacent points, and when counted mark them.
int DataBoard::CountUnMarkedLiberties( Pt *p )
{
	int row = p->GetRow();
	int col = p->GetCol();
	int libcount = 0;
	if( row > 0)
		libcount += _pBoard[row-1][col]._piece.MarkIfEmptyAndNotMarked();
	if( col > 0)
		libcount += _pBoard[row][col-1]._piece.MarkIfEmptyAndNotMarked();
	if( row < _maxsize - 1)
		libcount += _pBoard[row+1][col]._piece.MarkIfEmptyAndNotMarked();
	if( col < _maxsize - 1)
		libcount += _pBoard[row][col+1]._piece.MarkIfEmptyAndNotMarked();
	return libcount;
}

void DataBoard::ClearMarkedLiberties( Pt *p )
{
	int row = p->GetRow();
	int col = p->GetCol();
	if( row > 0)
		_pBoard[row-1][col]._piece.ClearMark();
	if( col > 0)
		_pBoard[row][col-1]._piece.ClearMark();
	if( row < _maxsize - 1)
		_pBoard[row+1][col]._piece.ClearMark();
	if( col < _maxsize - 1)
		_pBoard[row][col+1]._piece.ClearMark();
}

// -----------------------------------------------------------------

/**
// This text has to do with adding a move to the board and doing
//	incremental updates of the data structures.
//	The text is incomplete and I keep it here only as a record of
//	thinking about it.  I note that Fotland and others have noted on
//	computer-go-mailing-list that they spend alot of effort to doing
//	incremental updates and that it wasn't worth the complexity.
//	Fotland notes doing it from scratch (non-incremental) and that
//	that was sufficient.  So that is the way I do it... 11/96
	// The piece is added to any adjacent string of the same color.
	//	Adding the piece may cause two adjacent strings to merge.
	//	If the piece is not next to any of the same color, it creates
	//	a new GoString and joins it.
	// The algorithm is:
	//	1.	The piece initially is not in any GoString.
	// 	2.	Cycle through each adjacent point on the board,
	//		2a.	If the adjacent cell is the same color,
	//			join that point's GoString.
	//		2b.	For the remaining adjacent points of the same color,
	//			merge them with the piece's GoString.
	//		2c.	For any adjacent point not the same color, re-calculate
	//			it's liberty count.
	//	3.	Once the cycle completes, if the piece is not in a GoString,
	//			then it isolated.  Create a new GoString and join it.

	// Minimal Recalc method:
	// 1. Recalc the Cell's GoString's liberty count.
	// 2. Cycle through each adjacent cell and determine:
	// 3.	if( liberty count > 0 )	it's alive, but did it kill any strings?
	//			recalc liberty of each adjacent 

	// 2. If Cell's liberty count is 0 (just filled the last liberty), then:
	// 3.	cycle through each adjacent cell and determine:
	// 		If the stone is surrounded by his own men, the
	//			player is committing suicide.
	//		If all the opponent stones each have a liberty count > 1,
	//			player is committing suicide.
	// 4. If suicide, then
	//		if japanese rules, then takeback move... and reject.
	// 		else kill Cell's GoString and recalc liberties of adjacent
	//		strings.
	// 5. Else (not suicide), for each adjacent GoString, recalc liberty.
	//	
	// Now kill any GoStrings with no liberties. Add adjacent strings to
	//	a recalc list.
*/

// -----------------------------------------------------------------

BOOL DataBoard::GetPieceAtPos( Pt *p, PieceColor *pC ) const
{
	Coord row, col;
	p->GetRowCol( &row, &col);
	if( (row < 0 )
	||	(col < 0 )
	||	(row >= _maxsize )
	||	(col >= _maxsize ))
		return FALSE;

	*pC = _pBoard[row][col]._piece.GetColor();
	return TRUE;
}

BOOL DataBoard::MarkAsDead(Pt * position) {
  Coord row, col;
  position->GetRowCol( &row, &col);
  PieceColor pc =  _pBoard[row][col]._piece.GetColor();
  // sanity check
  if( col >= _maxsize || row >= _maxsize ||pc == PC_None)
    return FALSE;
  GoStringCollection *gsc;
  if (pc == PC_White)
    gsc = &_WhiteStrings;
  else
    gsc = &_BlackStrings; 
  //get attached string
  GoString * gs = gsc->Find(position);
  if (!gs)
    Fatal( ERR_PANIC );
  int c = gs->GetStoneCount();
  // fprintf(stderr,"%d stoned removed\n",c);
  if (pc == PC_White)
    _DeadWhite += c;
  else
    _DeadBlack +=c;
  vector<Piece> pi;
  gs->CopyToListAndClear(&pi);
  gsc->Remove(gs);
  pi.clear();
  _ReCalcAllGoStrings(pc, &pi);
  _CreateEnclosures();
  return TRUE;
}

/**
// This function is called to add one move to the board.
//	Return 1 if ok, 2 if illegal suicide, 3 if ko violation;
//		MOVE_OK
//		MOVE_ILLEGAL_SUICIDE
//		MOVE_KO_VIOLATION
*/
int	DataBoard::SetMove( const Move *pMv )
{
	int MoveResult = MOVE_OK;
	vector<Piece> TheDead;
	if( pMv->GetMoveType() == MT_Move )
	{
		Coord row, col;
		pMv->GetCoords( &row, &col);
		PieceColor color = pMv->GetColor();

		// sanity check
		if( col >= _maxsize || row >= _maxsize || _pBoard[row][col]._piece.GetColor() != PC_None)
		  return MOVE_ILLEGAL_SUICIDE;
		Assert( color != PC_None );

		// Ok, set the piece in the board.
		_pBoard[row][col]._piece.SetColor( color );


		// Update the board data structures...
		if( _ReCalcAllGoStrings ( color, &TheDead ) && (!_bSuicideIsOk) )
		{
			// illegal suicide
			MoveResult = MOVE_ILLEGAL_SUICIDE;

fail:
			// undo the move:
			_pBoard[row][col]._piece.SetColor( PC_None );

			// undo those that were killed too:
			vector<Piece>::iterator pi;
			for( pi = TheDead.begin(); pi != TheDead.end(); ++pi )
			{
				Coord		KilledRow	= (*pi).GetRow();
				Coord		KilledCol	= (*pi).GetCol();
				PieceColor	KilledColor	= (*pi).GetColor();

				// if the last move was suicide, don't add it back
				//	or it'll get killed again on recalc.
				if( row == KilledRow && col == KilledCol )
					continue;
				_pBoard[KilledRow][KilledCol]._piece.SetColor( KilledColor );
			}
			// Recalc GoStrings and Enclosures of the original board.
			vector<Piece> UnDead;
			if( _ReCalcAllGoStrings( color, &UnDead ) )
				Fatal(ERR_PANIC);	// can't suicide on move undo...
			Assert( UnDead.size() == 0 );	// no one should have died on undo.

			// Now that we have the occupied GoStrings, determine the Enclosures.
			_CreateEnclosures();

			return MoveResult;
		}

		// Check for Ko
#if USE_KOCHECK
/**
// This code assumes that the hash is unique for every board.  If the hash
//	algorithm is likely to give two different boards the same hash, then
//	this code will fail.  Why?  Not for a good reason: this code does not
//	keep each board and it's hash such that if a hash collision occurs then
//	the boards can be compared.  That would be the right thing to do...
//	However, we are now using the Zobrist algorithm for which a collision is
//	unlikely.
//	NB: If such a collision occurs, then the program will appear to be hung.
//	For instance, if wally plays a move and this code thinks it is a repeat
//	of a previous board, then a false KO will be detected. The move will
//	be sent back to wally who will simply repeat the same move, and the
//	process repeats ad infinitum.
// This really should be fixed...
*/

		U32BIT	BoardHashVal = _ComputeHashVal();
		if( _ViolatesKo( BoardHashVal ) )
		{
			// illegal ko, undo the move:
			MoveResult = MOVE_KO_VIOLATION;
			goto fail;
		}
		_SaveBoardHashVal( BoardHashVal );
#endif // USE_KOCHECK

		// Now that we have the occupied GoStrings, determine the Enclosures.
		_CreateEnclosures();
_ShowEnclosures(); // DIAGNOSTICS

	}
	// else	// passes

	// add move to history...
	_AllMoves.AddMove( pMv, &TheDead );

	return MoveResult;
}

void	DataBoard::TakeBackMove()
{
	// Get the last move and those who died from the history...
	Move LastMv;
	vector<Piece> LastVictims;
	_AllMoves.TakeBackMove( &LastMv, &LastVictims );

	// Clear the last move...
	_pBoard[LastMv.GetRow()][LastMv.GetCol()]._piece.SetColor( PC_None );

	// Add back dead....
	vector<Piece>::iterator pi;
	for( pi = LastVictims.begin(); pi != LastVictims.end(); ++pi )
		_pBoard[(*pi).GetRow()][(*pi).GetCol()]._piece.
			SetColor( (*pi).GetColor() );

	// Reset the board...
	vector<Piece> TheDead;
	Move Mv;
	BOOL bSuicided = _ReCalcAllGoStrings( Mv.GetColor(), &TheDead );
	Assert( bSuicided == 0);		// prior move should be stable
	Assert( TheDead.size() == 0 );	// prior move should be stable
}

void	DataBoard::GetBoard( Board_OptSpace *pBdOS, U8BIT * pBdSz ) const
{
	for( int row = 0; row < _maxsize; row++ ) {
		for( int col = 0; col < _maxsize; col++ ) {
			pBdOS->SetColorAtRowCol(
				row, col,
//				ConvertPlayerColorToDBColor( _pBoard[row][col]._piece.GetColor() )
				_pBoard[row][col]._piece.GetColor()
				);
            }
        }
	*pBdSz = _maxsize;
}
// -----------------------------------------------------------------

void DataBoard::_ClearStrings()
{
	// Clear the Cells references to the GoStrings and free the GoStrings.
	for( int row = 0; row < _maxsize; row++ ) 
		for( int col = 0; col < _maxsize; col++ ) 
			_pBoard[row][col]._pInStr = 0;
	_WhiteStrings.Flush();
	_BlackStrings.Flush();
}

void DataBoard::_CreateStrings()
{
	// Walk board and re-create strings.
	for( int row = 0; row < _maxsize; row++ ) {
		for( int col = 0; col < _maxsize; col++ ) {
			BoardCell * pCell = &_pBoard[row][col];
			PieceColor pc = pCell->_piece.GetColor();
			if( pc == PC_None )
				continue;
Assert(pCell->_pInStr==0);

			GoString * pAdjRowStr = 0;
			GoString * pAdjColStr = 0;

			if( (row > 0)
			&& (_pBoard[row-1][col]._piece.GetColor() == pc) )
			{
				pAdjRowStr = _pBoard[row-1][col]._pInStr;
				Assert( pAdjRowStr );
				pAdjRowStr->AddOccupiedCell( pCell );
			}


			if( (col > 0)
			&& (_pBoard[row][col-1]._piece.GetColor() == pc) )
			{
				// if the adjacent string in the row is not
				//	the same as the one in the column, then
				//	merge them.
				// NB: the adjacent string in the code above
				//	has just added this boardcell to it, so
				//	after merging it would be wrong to add the
				//	cell again because that would put it into
				//	string twice -- a fatal error.
				pAdjColStr = _pBoard[row][col-1]._pInStr;
				Assert( pAdjColStr );
                if( pAdjRowStr ) {
					if( pAdjRowStr != pAdjColStr )
					{
						pAdjRowStr->Absorb( pAdjColStr );
						delete pAdjColStr;

						// remove from string collection too:
						if( pCell->_piece.GetColor() == PC_Black )
							_BlackStrings.Remove( pAdjColStr );
						else
						if( pCell->_piece.GetColor() == PC_White )
							_WhiteStrings.Remove( pAdjColStr );

						pAdjColStr = pAdjRowStr;
					}
				}
				else
					pAdjColStr->AddOccupiedCell( pCell );
			}


			if( (pAdjRowStr==0) && (pAdjColStr==0) ) {
				// create standalone string
				GoString *pInStr = new GoString();
				Assert( pInStr );
				pInStr->AddOccupiedCell( pCell );

				// add to proper collection
				if( pCell->_piece.GetColor() == PC_Black )
					_BlackStrings.Add( pInStr );
				else
				if( pCell->_piece.GetColor() == PC_White )
					_WhiteStrings.Add( pInStr );
				else
					Fatal(ERR_ASSERT);
			}
		}
	}
}

//! Clear board's liberty count markers, then compute each string's liberties.
void DataBoard::_CalcLibertiesOfStrings()
{
	// remove all counting marks from the board.
	for( int row = 0; row < _maxsize; row++ )
		for( int col = 0; col < _maxsize; col++ )
			_pBoard[row][col]._piece.ClearMark();

	_BlackStrings.CalcLiberties(this);
	_WhiteStrings.CalcLiberties(this);
}


/**
// This function will properly remove strings from the board taking
//	into account who moved last.
//	It calculates all the strings liberties as needed.
//	in accordance with the input of Who Moved Last.
// Return True if WhoMovedLast committed suicide.  If WhoMovedLast
//	is PC_None (nobody - like on a takeback), then suicide can't occur.
*/
BOOL DataBoard::_RemoveDeadStrings( PieceColor WhoMovedLast, vector<Piece> *pTheDead )
{
	int TheirDead;
	int OurDead;

	_CalcLibertiesOfStrings();

	// Kill opponent first...

	TheirDead = 0;
	if( WhoMovedLast == PC_Black )
		TheirDead = _WhiteStrings.RemoveDead( pTheDead );
	else
		TheirDead = _BlackStrings.RemoveDead( pTheDead );

	if( TheirDead )
		_CalcLibertiesOfStrings();	// something died, so recalc.

	// then kill self...

	OurDead = 0;
	if( WhoMovedLast == PC_Black )
		OurDead = _BlackStrings.RemoveDead( pTheDead );
	else
		OurDead = _WhiteStrings.RemoveDead( pTheDead );

// if( TheirDead )
// cout << "DataBoard: Killed " << TheirDead << " of theirs." << endl;
// if( OurDead )
// cout << "DataBoard: Killed " << OurDead << " of mine." << endl;

 	// Determine if WhoMovedLast committed suicide.
	if( (WhoMovedLast != PC_None) && (TheirDead == 0) && OurDead)
		return TRUE;	// suicide
	return FALSE;
}

// -----------------------------------------------------------------

//! Return FALSE if suicide, TRUE otherwise.
BOOL DataBoard::_ReCalcAllGoStrings( PieceColor WhoMovedLast, vector<Piece> *pTheDead)
{
	/**
	// A: (Clear all the data structures except for the stones on the board).
	//  1. Remove all the GoStrings and clear all Cell references to GoStrings.
	//  2. Clear the GoStringCollections.
	// B: (Add Cells to GoStrings and the GoStrings to the GoStringCollections)
	//  1. Walk board and for each occupied cell:
	//  2.	if previous or upper adjacent cell is occupied with same color:
	//  3.		add cell to previously created GoString
	//  4.		else: create new GoString and add cell to it.
	//	5.		Add GoString to proper GoStringCollection
	// C: (Determine Liberties)
	//  1. For each GoString
	//	2.	For each Cell
	//	3.		Count each unmarked empty adjacent Cell.
	//	3.		mark the empty adjacent Cell.
	//	4	Set the GoStrings liberty count.
	//	5.	For each Cell
	//	6.		Clear each marked adjacent cell.
	// D: (Kill those GoStrings that have no air).
	//  1. Walk WhoMovedLast's opponent's GoStrings and kill each with
	//		libcount==0.
	//  2. If at least one GoString killed, Recalc Liberties. (step C)
	//  3. Then Walk WhoMovedLast's GoStrings and kill each with libcount==0.
	//  4. If at least one GoString killed, Recalc Liberties. (step C)
	//	5. If only WhoMovedLast's stones died, then report a suicide.
	*/


		// Step A:
	_ClearStrings();
		// Step B:
	_CreateStrings();
		// Step C & D:
	return _RemoveDeadStrings( WhoMovedLast, pTheDead );
}


// -----------------------------------------------------------------


void DataBoard::_DumpEnclosures()
{
	vector<Enclosure *>::iterator ppIter;

	// delete each enclosure in each of the lists.
	for( ppIter = _Enclosures.begin(); ppIter != _Enclosures.end(); ppIter++ )
		delete (*ppIter);
	_Enclosures.clear();
}

void DataBoard::_CreateEnclosures()
{

#if ALGORITHM
	/**
	// For every empty and unmarked BoardCell
	//	mark the BoardCell
	//	Create a Empty GoString
	//	add the BoardCell to the Empty GoString
	//	FloodFill to bordering unmarked BoardCells
	//		For cell hit:
	//		if BoardCell is empty, add to Empty GoString and mark it
	//		else if FirstColorHit==PC_None
	//			FirstColorHit = BoardCell.piece.GetColor()
	//			mark enclosure as territory of FirstColorHit
	//			add GoString to Border GoString list
	//			mark each cell of the bordering GoString
	//		else if BoardCell.piece.GetColor == FirstColorHit
	//			add GoString to Border GoString list
	//			mark each cell of the bordering GoString
	//		else // not the same as FirstColorHit
	//			mark enclosure as mixed ownership
	//			add GoString to Border GoString list (mixed list)
	//			mark each cell of the bordering GoString
	//		end flood fill.
	//			Unmark all the occupied GoStrings cells
	*/
#endif

	_DumpEnclosures();

	Coord r, c;
	// for every empty unmarked cell:
	for( r = 0; r < _maxsize; r++ )
	for( c = 0; c < _maxsize; c++ )
	{
		BoardCell *pBC = &_pBoard[r][c];

		// empty and unmarked...
		if( (pBC->_piece.GetColor() == PC_None) && (!pBC->_piece.IsMarked()) )
		{
			// create an enclosure...
			Enclosure * pEnclosure = new Enclosure();

			_FloodFillToBorders(pEnclosure, r, c);

			// add to enclosure pool...
			//	Note that the enclosures are keyed on their GetBorderColor()
			//	pEnclosure->GetBorderColor()
			_Enclosures.push_back( pEnclosure );

			// clear the marks in the GoString borders.
			pEnclosure->UnMarkBorderCells();
		}
	}

	// clear all marks...
	for( r = 0; r < _maxsize; r++ )
	for( c = 0; c < _maxsize; c++ )
	{
		_pBoard[r][c]._piece.ClearMark();
	}

}


//! DIAGNOSTICS
void DataBoard::_ShowEnclosures()
{
#ifdef DIAGNOSTICS
	vector<Enclosure *>::iterator ppIter;

	cout << "Enclosures " << _Enclosures.size() << endl;
	for( ppIter = _Enclosures.begin(); ppIter != _Enclosures.end();ppIter++)
	{
		PieceColor pc = (*ppIter)->GetBorderColor();
		cout << "\t" << 
			(pc==PC_Black?"Black":
			 (pc==PC_White?"White":
			  (pc==PC_None?"None ":
			   (pc==(PC_Black|PC_White)?"Mixed":
			     "??"))));
		cout << ": size " << (*ppIter)->Size() << ", border strings " <<
			(*ppIter)->BorderStringCount() << endl;
		Assert( pc != PC_None );
	}
#endif // DIAGNOSTICS
}


//! Return TRUE is score is final, else it is interim.
BOOL DataBoard::_ScoreGame( U32BIT &black, U32BIT &white, U32BIT &unsettled)
{
	vector<Enclosure *>::iterator ppIter;

	// U32BIT black = 0, white = 0, none = 0, unsettled = 0;
	U32BIT none = 0;
	black = white = unsettled = 0;
	BOOL bAllAreasAreSizeOne=TRUE;
	for( ppIter = _Enclosures.begin(); ppIter != _Enclosures.end();ppIter++)
	{
		switch( (*ppIter)->GetBorderColor() )
		{
		case PC_Black:	black += (*ppIter)->Size();	break;
		case PC_White:	white += (*ppIter)->Size();	break;
		case PC_None:	none += (*ppIter)->Size();	break;
		case PC_Black | PC_White:
						unsettled += (*ppIter)->Size();	break;
		default:	Assert(0);
		}
		if( (*ppIter)->Size() > 1 )
			bAllAreasAreSizeOne = FALSE;
	}
	//this can be true in one case: empty board Assert( none == 0 );

//	if( bAllAreasAreSizeOne )
//		cout << "Final Score: black " << black << ", white " << white
//			<< ", unsettled " << unsettled << endl;
//	else
//		cout << "Interim Score: black " << black << ", white " << white
//			<< ", unsettled " << unsettled << endl;

	// NOTE: if each unsettled is of size one, then each mixed enclosure should
	//	be checked to see if it is the single liberty for each bordering string,
	//	which would make it a stable seki point.

	return bAllAreasAreSizeOne;
}

void DataBoard::_FloodFillToBorders( Enclosure * pEnclosure, Coord row, Coord col)
{
	BoardCell *pBC = &_pBoard[row][col];

	// skip if already marked...
	if( pBC->_piece.IsMarked() )
		return;

	// if empty, then mark and add to empty list...
	if( pBC->_piece.GetColor() == PC_None )
	{
		pEnclosure->AddEmptyCell( pBC );
		pBC->_piece.SetMark();

		// Floodfill to next adjacent cell.
		if( row > 0 )
			_FloodFillToBorders( pEnclosure,	row-1,	col);
		if( col > 0 )
			_FloodFillToBorders( pEnclosure,	row,	col-1);
		if( row < _maxsize - 1 )
			_FloodFillToBorders( pEnclosure,	row+1,	col);
		if( col < _maxsize - 1 )
			_FloodFillToBorders( pEnclosure,	row,	col+1);
	}
	else// Hit an occpied cell.
		// if first occupied hit, then
		//	set 1st color, set border color, add string to border list, and
		//	mark border cells.
	if( pEnclosure->GetFirstColorHit() == PC_None )
	{
		pEnclosure->SetFirstColorHit( pBC->_piece.GetColor() );
		pEnclosure->SetBorderColor( pBC->_piece.GetColor() );
		pEnclosure->AddBorder( pBC->_pInStr );
	}
	else// Hit another occupied cell...
		// if same as first color hit, then
		//	add string to border list and mark border cells.
	if( pEnclosure->GetFirstColorHit() == pBC->_piece.GetColor() )
	{
		pEnclosure->AddBorder( pBC->_pInStr );
	}
	else// Hit another occupied cell...
		// Must be a different color then first hit,
		//	reset border color to be mixed,
		//	add string to border list and mark border cells.
	{
		pEnclosure->SetBorderColor( PC_White | PC_Black );	// mixed...
		pEnclosure->AddBorder( pBC->_pInStr );
	}

}


// -----------------------------------------------------------------

// Ko detection 
#ifdef ALGORITHM
	/** Ko Detection
	// Hash the board. (Uses Zobrist).
	// Check if hash already in hash table.
	// If so, it ViolatesKo, else not
	*/
#endif


BOOL DataBoard::_ViolatesKo( U32BIT HashVal )
{
	if( _bSimpleKo )
	{
		// just check the move before last.
		int n = _BoardHashVals.size() -1;
		if( n > 0 )
//			if( _BoardHashVals.at(n-1) == HashVal )
			if( _BoardHashVals[n-1] == HashVal )
			{
				_pIO->Info("DataBoard: found hashval %lx", HashVal );
				return TRUE;	// violates ko
			}
	}
	else
	{
		if( find(_BoardHashVals.begin(),_BoardHashVals.end(), HashVal)
			!= _BoardHashVals.end() )
		{
			_pIO->Info("DataBoard: found hashval %lx", HashVal );
			return TRUE;	// violates ko
		}
	}
	return FALSE;	// not in table, so doesn't violate ko.
}


void DataBoard::_SaveBoardHashVal( U32BIT HashVal )
{
  //_pIO->Info("DataBoard: hashval %lx", HashVal );
	_BoardHashVals.push_back( HashVal );
}


/**
// Compute the board's hash value from scratch.
//	Since the hash involves XOR it could be done very fast incrementally.
//	But as this code is not called when pieces are removed, it must
//	compute the board from scratch, rather than incrementally.
//	The updated hash value is returned.
//	Hashing is done using the Zobrist algorithm.
*/
U32BIT	DataBoard::_ComputeHashVal()
{
	//
	// Create the random hash values for each point.
	//
	static U32BIT blackHashVals[ MAX_COLS ][ MAX_ROWS ];
	static U32BIT whiteHashVals[ MAX_COLS ][ MAX_ROWS ];
	static BOOL bFirstTime = FALSE;
	if( !bFirstTime )
	{
		bFirstTime = TRUE;
		for( int r = 0; r < MAX_COLS; r++ )
		for( int c = 0; c < MAX_ROWS; c++ )
		{
		  
			blackHashVals[r][c] = rnd_FullRange();
			whiteHashVals[r][c] = rnd_FullRange();
			if (whiteHashVals[r][c]==0 ||blackHashVals[r][c]==0)
			  exit(1);
		}
	}

	//
	// Recompute hash for current board.
	//
	U32BIT HashVal = 0;
	for( int r = 0; r < _maxsize; r++ )
	for( int c = 0; c < _maxsize; c++ )
	{
	  if( _pBoard[r][c]._piece.GetColor() == PC_Black )
	    {
	    HashVal ^= blackHashVals[r][c];

	  }
	  else
	    if( _pBoard[r][c]._piece.GetColor() == PC_White ) {
	      HashVal ^= whiteHashVals[r][c];
	    }
	}

	return HashVal;
}

// -----------------------------------------------------------------

#if 0
// ***************************************
ErrId	DataBoard::EstimateScore( int *pWhiteScore, int *pBlackScore )
{
	Assert( pBlackScore && pWhiteScore );

	int komi = 0;
	int WhiteArea = 0;
	int BlackArea = 0;

	*pWhiteScore = WhiteArea + komi - _WhiteCaptured;
	*pBlackScore = BlackArea - _BlackCaptured;

	return ERR_NONE;
}
#endif


void DataBoard::GetNthMove( Move *pMv, vector<Piece> *pKilled, int n )
{
	_AllMoves.GetNthMove( pMv, pKilled, n );
}
