//! databd.h
//!	A bunch of classes used to calculate go information and store it as
//!	a go board.
//
// The contents of this file are subject to
//	the OpenGo Public License Version 2.0 (the "License"); you may not
//	use this file except in compliance with the License. You may obtain
//	a copy of the License at http://www.inventivity.com/OpenGo
//
// Software distributed under the License is distributed on an "AS IS" basis,
//	WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
//	for the specific language governing rights and limitations under
//	the License.
// 
// Copyright:
//      Matthieu Nottale        copyright (c) 2005
//	Jeffrey Greenberg	copyright (c) 1996-2002
// 
// Contributor(s):
//	Carl Staelin	v5.2	linux compile issues
//
//==========================================================================


#ifndef INC_DATABD
#define INC_DATABD

#include "basetype.h"
#include "deps.h"
#include "fatal.h"
#include "move.h"
#include "iobase.h"
#include "bdspace.h"
//#include "games.h"	// for GetBoard() member function.

class GoString;	// forward...
class DataBoard;	// forward...


//! A cell of the board.
class BoardCell
{
	public:
		BoardCell()
			{
				_pInStr = 0;
			};
		~BoardCell()
			{
				_pInStr = 0;
			};
	private:
		BoardCell( BoardCell &)
			{
				Assert(0);
			};
		BoardCell & operator=(BoardCell &rhs)
			{
				Assert(0);
				return rhs;
			};
	public:
		// These are items that speed up access given a boardcell...
		Piece		_piece;			//! what piece is here
		GoString *	_pInStr;		//! member of some string.
};

// **********************************************************

//! A contiguous set of stones aka string or group.
class GoString
{
	public:
		GoString();
		~GoString();
	private:
		GoString( GoString &)
			{
				Assert(0);
			};
		GoString & operator=( GoString &rhs)
			{
				Assert(0);
				return rhs;
			};

	private:
		typedef list< BoardCell * >	ListOf_pCells;
	public:
			//! put cell into string and point string into cell.
		void AddOccupiedCell( BoardCell *pCell );
		void AddEmptyCell( BoardCell *pCell );

		void MarkAll();
		void UnMarkAll();

		void Absorb( GoString *pOther);

			//! Calculates liberties for string.
			//! 	Board is marked and cleared in the process...
		int CalcLiberties( DataBoard *pBoard );

		int GetLibertyCount()
			{
				return _LibertyCount;
			};
		int GetStoneCount()
			{
				return _Cells.size();
			};

			//! Copy a cell to the DeadList, then clear it's
			//!	references to the GoString and make the cell empty.
		void CopyToListAndClear( vector<Piece> * pList );
			//! Clear the references to the GoString.
		void StopCellsFromUsingMe();

			//! debugging functions
		void CheckForRepeats();
		void Validate();
		
		//return true if the string contains given element
		BOOL Contains(Pt * pt);
	private:
		int				_LibertyCount;	//! liberties for the string
		ListOf_pCells	_Cells;			//! cells that compose this string
//		ListOf_pCells	_OpponentCells	// cells of opponent that might
//										//	be affected by our death.
//		ListOf_pCells	_Liberties		// cells that comprise our liberties.
//		Shape			_ShapeType;		// Description of this shape, eg. eye
};

// **********************************************************

//! A collection of GoString
class GoStringCollection
{
	public:
		GoStringCollection();
		~GoStringCollection();
		
	private:
		typedef deque<GoString *> ListOf_GoStrings;
	public:
		//! remove and delete each GoString stored in the collection.
		void Flush();

		int CalcLiberties( DataBoard *pBoard );

		int GetLibertyCount();

		int RemoveDead( vector<Piece> * pTheDead );
		
		BOOL Remove( GoString *pVictim );
		
		//returns the gostring containing the given point
		GoString * Find(Pt * point);

		void Add( GoString *pGS );

	private:
		ListOf_GoStrings	_Strings;
		int					_TotalLiberties;	// liberties of this collection
};


// **********************************************************
/** Enclosure & Eyes (Scoring & Stable-Shapes & Connections)
// 
//	*Enclosure*
// An Enclosure is defined by a region of empty cells that are surrounded
// by one or more adjacent GoStrings all of the same color.
// An enclosure has no leaks, the GoStrings being separated at most by
// a diagonal link. In general, an enclosure may therefore be cuttable.
//
//	*Eye*
// An Eye is an enclosure that either has or can be reduced to (by fighting)
// a single GoString bounding one empty cell, a ponuki.
// Go literature describes many such enclosures which are known to be
// reducable to ponuki or not.  The enclosure should be capable of answering
// whether it can become ponuki or not, whether it cannot tell.  Questions
// about an enclosure are often referred to in the literature as issues
// about "shape".
//
//	*Scoring*
// Enclosures, being built from GoStrings, are almost as fundamental and
// are neccesary for scoring.  Once all the empty points are a member of
// an enclosure, then when all the enclosures can describe whether they
// are stable, the game is over and can be scored.
//
//	*Double-Eyes*
// An enclosure that is reducable to ponuki and which shares GoStrings
// with another eye is stable and known as a double-eye.  Such a structure
// cannot be destroyed by definition (as a result of Go rules).
//
//	*Connections*
// This definition of an enclosure needs to be kept strict because it
// separates out a class of moves that connect loosely connected GoStrings
// such that they can become part of an Enclosure (hopefully one reducable
// to ponuki or a double-eye.  Such moves are of importance.
// In addition a slightly less important class of moves are those that
// bring GoStrings close enough or give them security enough to become
// loosely connected on the way to becoming full-blown, stable Enclosures
// and double eyes.
//
// 	*Major Problem With This Algorithm*
// 	1. Note that this algorithm FAILS when it surrounds a one or more GoStrings
// 	which could be killed.  In this case, the algorithm will not create an
// 	Enclosure, and there not Eye knowlege: ie is reducable to ponuki...
// 	perhaps Enclosures and Eye need to be merged... that is there are
// 	enclosures that can include other groups by flooding around them which in
// 	turn might have enclosures within them.  Determining the status of each
// 	becomes non-trivial in that it would depend on who moves first and in what
// 	enclosure perhaps...
//	2. Nonetheless, the definition of Enclosure is suitable for determining
//	game-end in that until a structure that has region within regions can be
//	resolved into an Enclosure, that game is not over.
//
// GoStrings can be members of one or more enclosures
*/


/**	Enclosur*
// An Enclosure is defined by a region of empty cells that are surrounded
// by one or more adjacent GoStrings all of the same color.
// An enclosure has no leaks, the GoStrings being separated at most by
// a diagonal link. In general, an enclosure may therefore be cuttable.
*/
class Enclosure
{
	public:
		Enclosure()
			{
				_pEmptyCells = 0;
				_FirstColorHit = PC_None;
				_BorderColor = PC_None;
			};
		~Enclosure()
			{
				if( _pEmptyCells )
					delete _pEmptyCells;
				_pEmptyCells = 0;
			};

	private:
		Enclosure & operator=( Enclosure & rhs )
			{
				Fatal(ERR_PANIC);
				return rhs;
			};
		Enclosure( Enclosure &e )
			{
				Fatal(ERR_PANIC);
			};

	private:
		typedef vector<GoString *> ListOf_GoStrings;

	public:

#if LATER
		//! Status()
		//!	Member function does the following:
		//! 	returns one of: IsSinglePoint, IsReducibleToSinglePoint, Unknown.
		enum StatusValue = { Unknown, IsSinglePoint, IsReducibleToSinglePoint };

		StatusValue Status()
			{
				return Unknown;
			};
#endif

		void AddEmptyCell( BoardCell *pBC )
			{
				if( !_pEmptyCells )
					_pEmptyCells = new GoString();
// where does it get freed?
				_pEmptyCells->AddEmptyCell(pBC);
			};

		void AddBorder( GoString *pGS )
			{
				_BorderStrings.push_back(pGS);
				pGS->MarkAll();
			};

		void UnMarkBorderCells()
			{
				ListOf_GoStrings::iterator ppGS;
				for( ppGS = _BorderStrings.begin();
					ppGS != _BorderStrings.end();
					++ppGS
					)
						(*ppGS)->UnMarkAll();
			};

		PieceColor GetFirstColorHit()
			{
				return _FirstColorHit;
			};

		void SetFirstColorHit( PieceColor color)
			{
				_FirstColorHit = color;
			};

		PieceColor GetBorderColor()
			{
				return _BorderColor;
			};

		void SetBorderColor( PieceColor color)
			{
				_BorderColor = color;
			};

			// number of empty cells
		int Size()
			{
				return _pEmptyCells->GetStoneCount();
			};

			// number of border strings
		int BorderStringCount()
			{
				return _BorderStrings.size();
			};


		// IsIn( Pt p );		// in anywhere
		// IsInInside( Pt p );	// in empty area
		// IsInBorder( Pt p );	// in border area
		//	(What about GoString entirely inside an enclosure??)
		//

	private:

		// Data Structures
		// 	list of empty BoardCells.
		// 		* number of cells (eye count or eye size).
		// 	list of GoStrings of same color that border this eye.
		// 		* number of GoStrings that surround this eye.

		GoString	*		_pEmptyCells;
		ListOf_GoStrings	_BorderStrings;
		PieceColor			_BorderColor;	// Color of Border strings.
											//	PC_Black | PC_White == mixed color
		PieceColor			_FirstColorHit;	// used in flood fill
};


// class Linkages
//{
//	typedef list<GoString *>	Link;
//	Link	LooseLink;
//	int		type;	// knight..., etc...
//	static int		property[] = Cuttable,	// indexed by type
//};


// **********************************************************
//! HistoryElement is a pair of a move and pieces it killed.
class HistoryElement
{
	public:
		HistoryElement()
//			: _mv(mv), _killed(*pKilled)
			{};
		HistoryElement( const Move mv, const vector<Piece> * pKilled )
			: _mv(mv), _killed(*pKilled)
			{};
		~HistoryElement() {};
	public:
		void CountDead( int *pWhite, int *pBlack )
			{
				vector<Piece>::iterator p;
				for( p = _killed.begin();p!=_killed.end();p++)
				{
					if( (*p).GetColor() == PC_White )
						++(*pWhite);
					else
					if( (*p).GetColor() == PC_Black )
						++(*pBlack);
					else
						Fatal(ERR_ASSERT);
				}
			}
	public:
		Move 			_mv;
		vector<Piece>	_killed;
};


//! A collection of HistoryElement
//!	All the moves in the game.
class History
{
	public:
		History() {};
		~History() {};

	private:
		typedef vector<HistoryElement> ListOf_HistoryElements;

	public:
		void AddMove( const Move *pMv, const vector<Piece> *pKilled  )
			{

				#ifdef _DEBUG
					// validate: don't kill spaces
					vector<Piece>::const_iterator p;
					for( p = pKilled->begin();p!=pKilled->end();p++)
					{
						if( (*p).GetColor() == PC_White )
							;
						else
						if( (*p).GetColor() == PC_Black )
							;
						else
							Fatal(ERR_ASSERT);
					}
				#endif

				_History.push_back( HistoryElement(*pMv, pKilled) );
			}

		void TakeBackMove( Move *pMv, vector<Piece> *pKilled )
			{
				if( _History.size() )
				{
					HistoryElement HistElem = _History.back();
					*pMv = HistElem._mv;
					*pKilled = HistElem._killed;
					_History.pop_back();
				}
			}

		// n <= 0 will get last move...
		void GetNthMove( Move *pMv, vector<Piece> *pKilled, int n=0 )
			{
				if( _History.size() )
				{
					if( (n <= 0) || (n >= (int)_History.size()) )
						n = _History.size() -1;	// get last

					HistoryElement HistElem = _History[n];
					*pMv = HistElem._mv;
					*pKilled = HistElem._killed;
				}
			}

		void GetCapturedCount( int *pWhite, int *pBlack )
			{
				*pWhite = *pBlack = 0;
				ListOf_HistoryElements::iterator p;
				for( p = _History.begin(); p != _History.end(); p++ )
					(*p).CountDead(pWhite,pBlack);
			}

		S32BIT GetMoveCount() const
			{
				return _History.size();
			}

	private:
		ListOf_HistoryElements	_History;
};

// **********************************************************
// DataBoard

#define MAX_COLS	19	//! maximum DataBoard col size
#define MAX_ROWS	19	//! maximum DataBoard row size

#define	MOVE_OK					1	//! good move
#define MOVE_ILLEGAL_SUICIDE	2	//! illegal suicide!
#define MOVE_KO_VIOLATION		3	//! ko violation!


/** DataBoard
// The DataBoard class describes all sorts of relationships in the current
//	board.
//	Currently:
//		Stone positions on the board
//		Stone membership in what string
//		Move History (each move has a list of stones killed as a result)
//		GoStrings
//			liberty count
//	Later:
//		GoStrings
//			absolutely alive (ie 2+eyes)
//	The DataBoard is updated via
//		SetMove(mv) which updates the board and removes dead stones.
//		TakebackMove() removes last move from history.
//		UpdateBoard() which recomputes the data structures from the given board.
//	Results can be obtained via:
//		GetBoard() which returns the current board Board_optspace
// =============
//
// The DataBoard goes through different phases of analysis:
//	1: a piece is added to the board
//	2: Create Player GoStrings (White & Black collections of GoStrings
//		occupied BoardCells point to GoStrings
//		GoStrings point to occupied BoardCells
//		Dead strings are removed
//		Suicide is detected if the rules prevent it and the move rejected.
//		If the move is ok, the move and the stones that were killed are stored
//			in the move history list.
//	3: Create Enclosures: Empty regions owned by White, Black, or are Shared.
//		a: create enclosure:
//			empty BoardCells are put into different kinds of Empty lists.
//			The bordering GoStrings are noted.
//		b: Determine stability of enclosures:
//			Classify enclosures: single eye, multiple eye, seki, or open.
//			Single eyes can be reduced to ponuki, killable, unknown.
//			Multiple eyes are inherently stable
//			Seki are stable for this player if they don't move there first.
//			Open space requires more reading.
//		c: Perhaps other analysis to ascertain stability of
//			unclassified single eyes and the open space strings.
//	===
//	At this point strategy comes in now that a tactical analysis has
//	been done. (It may be possible to determine single eye status somehow??)
//	For example, single-eyes can be joined to form multiple eyes, strings can
//	be killed thus making more empty space that might be stable, shared
//	open space can be occupied, strings can be connected, etc.  This might be
//	where gp could be applied.
//
// ==============
*/

class DataBoard
{

	public:
			/**
			// When board is constructed it is told:
			//	1. Whether to allow suicide moves or not
			//	2. Whether Ko is simple (disallow repeat of last move) or
			//		not (SuperKo - diallow any previously seen board).
			//	3. board size
			//	4. io controller for debug output.
			*/
		DataBoard( BOOL bSuicideIsOk, BOOL bSimpleKo, Coord Size, IOBase *pIO );
		~DataBoard();
	private:
		// prevent this usage...
		DataBoard( DataBoard & r) {Assert(0); };
		DataBoard() {Assert(0); };

	public:
		// Main interfaces to the engine:

		//! Add move to board, update data structures and save in history...
		//!	return 1 if ok, 2 if illegal suicide, 3 if ko violation
		int	SetMove( const Move *pMv );

        void SetBoardLoc(int row, int col, char Value)
            {
            _pBoard[row][col]._piece.SetColor( Value );
            //_pBoard[row][col] = Value;
            };

        void    SetBoardSize( Coord Size ) { _maxsize = Size; };

    	//! Check what piece is at a given position.
		//! 	return False if p is off the board.
		BOOL GetPieceAtPos( Pt *p, PieceColor *pC ) const;

		BOOL IsSuicideOk() const
		{
			return _bSuicideIsOk;
		}

			//! Take back last move from history...
		void	TakeBackMove();
		void GetNthMove( Move *pMv, vector<Piece> *pKilled, int n );

			//! Obtain the current board position stored in DataBoard.
		void	GetBoard( Board_OptSpace * pBoard, U8BIT *pBdSz ) const;
		void	GetBoardSize( Coord *pSize ) const
		{
			*pSize = _maxsize;
		}

			//! Get current total capture counts for each player.
		void GetCapturedCount( int *pWhite, int *pBlack )
		{
		  _AllMoves.GetCapturedCount( pWhite, pBlack );
		  (*pWhite) += _DeadWhite;
		  (*pBlack) += _DeadBlack;
		}

			//! What move is this?
		S32BIT GetMoveCount() const
		{
			return _AllMoves.GetMoveCount();
		}
		
		// remove this stone and all attached stones and add them to the prisoners
		BOOL MarkAsDead(Pt * position); 
	public:
		// These are really for the GoString class which should
		//	be a friend class it seems... but I'm not comfortable with
		//	that just yet.
			//! Count & mark point's unmarked liberties.
		int CountUnMarkedLiberties( Pt *p );
			//! Clear any marked liberties.
		void ClearMarkedLiberties( Pt *p );


	private:
		Coord			_maxsize;
		BoardCell		_pBoard[MAX_ROWS][MAX_COLS];
		IOBase *		_pIO;//! debugging io
	private:
		BOOL			_bSuicideIsOk;	//! affects whether move is ok or not.
		BOOL			_bSimpleKo;		//! Ko is simple or super.
	private:
		// These are operations on a board that are applied after a move.
		//	They are used to: 1. find/create GoStrings, 2. determine if the
		//	move killed any opponent strings, 3. determine if suicide is the
		//	result.
		void			_ClearStrings();
		void			_CreateStrings();
		void			_CalcLibertiesOfStrings();
		BOOL			_RemoveDeadStrings( PieceColor WhoMovedLast, vector<Piece> *pTheDead );
		BOOL			_ReCalcAllGoStrings( PieceColor WhoMovedLast, vector<Piece> *pTheDead);


	private:
		GoStringCollection	_WhiteStrings;
		GoStringCollection	_BlackStrings;
		int                     _DeadBlack,_DeadWhite;
	private:   
		History 		_AllMoves;	//! list of all moves made and stones taken

		// Ko
	private:
		BOOL			_ViolatesKo( U32BIT HashVal );
		void			_SaveBoardHashVal( U32BIT HashVal );
		U32BIT			_ComputeHashVal();
	private:
		vector<U32BIT>	_BoardHashVals;	//! previous board hash values.

//	// LATER
//	private:
//
//		// Lists of points:
//		PieceList *		_pEmptyPieces;		// all unoccupied points
//
//		// Scoring data
//		int				Score(int &whoWon, int &white, int &black);
//		int				_WhiteCaptured;		// count of captured white stones
//		int				_BlackCaptured;		// count of captured black stones


	//
	// Enclosure stuff
	//

private:

	void _CreateEnclosures();
	void _DumpEnclosures();

public:
	void _ShowEnclosures();	// debug

private:
	void _FloodFillToBorders( Enclosure * pEnclosure, Coord row, Coord col);
private:
	vector<Enclosure *> _Enclosures;


	// Scoring...
public:
	// return true if final
//	BOOL _ScoreGame();
	BOOL _ScoreGame( U32BIT &black, U32BIT &white, U32BIT &mixed);




/**
//	// For GP
// public:
//		// Is either in or boundary of enclosure:
//	IsInAnyEnclosure( Pt p);
//		// Is inside boundary of given enclosure:
//	IsEmptyCellInEnclosure( Pt p, Pt DefinesEnclosure);
//		// Is in boundary of given enclosure:
//	IsNonEmptyCellInBoundaryOfEnclosure( Pt p, Pt DefinesEnclosure);
//
//		// Is Enclosure part of double-eye (return false if not or unknown)
//	IsDoubleEye( Pt p );
//		// Is Single Eye (return false if not or unknown)
//	IsSingleEye( Pt p );
//		// Is reducible to Single Eye (return false if not or unknown)
//	IsReducibleToSingleEye( Pt p );
//
//
//	// For Scoring:
// public:
//		// ???
//		// If all points are enclosed and stable, and if playing only
//		// fills in a single point eye.
//		// If all points are enclosed and stable, then the game is over.
//		// 	How to handle seki? If point is single shared liberty of both
//		// 		adjacent GoStrings.
//		// 	If an Enclosure is enveloped by another and is not a DoubleEye,
//		// 		then it is dead unless Ko prevents it.
//		// 		How to figure Ko into this?
//	Bool AreAllEmptyPointsEnclosedAndStable();
//
//		// Then get the score...
//	GetTerritory( int Rules, int *pBlack, int *pWhite );
*/
};


#endif // INC_DATABD
